__package__ = 'pyscisci'
__title__ = 'pyscisci: A python package for the science of science'
__description__ = 'Lets study science!'

__copyright__ = '2021, Gates, A.J.'

__author__ = """\n""".join([
    'Alexander J Gates <ajgates42@gmail.com>'
])

__version__ = '0.8.11'
__release__ = '0.8.11'
